# ast.py
#
# Copyright (C) 2009, Codethink Ltd.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#
# Author:
#       John Carr <john.carr@unrouted.co.uk>

from . import endpoint

class Node(object):
    __slots__ = ()

    def get_string(self):
        raise NotImplementedError


class Expression(Node):
    __slots__ = ()


class WhereFragment(Node):
    __slots__ = ()


class SelectFragment(Node):
    __slots__ = ()


class Symbol(Expression):
     __slots__ = ()


class Select(SelectFragment):

    __slots__ = ("var", )

    def __init__(self, var):
        self.var = var

    def get_name(self):
        return self.var.name

    def get_string(self):
        return self.var.get_string()


class SelectAs(SelectFragment):
    __slots__ = ("expr", "alias")

    def __init__(self, expr, alias):
        if not isinstance(alias, Variable):
            raise ValueError("Can only select as a variable")
        self.expr = expr
        self.alias = alias

    def get_name(self):
        return self.alias.name

    def get_string(self):
        return "%s AS %s" % (self.expr.get_string(), self.alias.name)


class Function(Expression):
    __slots__ = ("function", "args")

    def __init__(self, function, *args):
        for arg in args:
            if not isinstance(arg, Symbol):
                 raise ValueError("Function can only accept symbols as arguments")

        self.function = function
        self.args = args

    def get_string(self):
        return "%s(%s)" % (self.function, ",".join(a.get_string() for a in self.args))


class Variable(Symbol):
    __slots__ = ("name", )

    def __init__(self, name):
        self.name = name

    def get_string(self):
        return "?" + self.name


class Constant(Symbol):
    __slots__ = ("val", )

    def __init__(self, val):
        self.val = val

    def _escape_string(self, val):
        val = val.decode("utf-8")
        val = val.replace("'", "\\'")
        val = val.replace("\"", "\\\"")
        return val

    def get_string(self):
        if isinstance(self.val, int):
            return str(self.val)
        else:
            return "\"%s\"" % self._escape_string(self.val)


class ConstantResource(Constant):
    __slots__ = ("val", )

    def __init__(self, val):
        self.val = val

    def get_string(self):
        return "<%s>" % self.val


class Relation(WhereFragment):
    __slots__ = ("subject", "predicate", "object")

    def __init__(self, subject, pred, object):
        self.subject = subject
        if isinstance(pred, basestring):
            pred = PrefixedName(pred)
        self.predicate = pred
        self.object = object

    def get_string(self):
        return "%s %s %s" % (self.subject.get_string(), self.predicate.get_string(), self.object.get_string())


class RelationSet(WhereFragment):
    __slots__ = ("subject", "type", "relations")

    def __init__(self, subject, type=None):
        self.subject = subject
        self.type = type
        self.relations = []

    def add(self, predicate, object):
        self.relations.append(Relation(self.subject, predicate, object))

    def get_string(self):
        first = True
        results = self.subject.get_string() + " "
        if self.type != None:
            results = results + "a " + self.type.get_string()
            first = False

        for r in self.relations:
            if first:
                first = False
            else:
                results = results + " ; "

            results = results + r.predicate.get_string() + " " + r.object.get_string ()


        return results


class Filter(WhereFragment):
    __slots__ = ("var", "test", "badger")

    def __init__(self, var, test, badger):
        self.var = var
        self.test = test
        self.badger = badger

    def get_string(self):
        return "FILTER (%s %s %s)" % (self.var.get_string, test, self.badger.get_string())


class Optional(WhereFragment):
    __slots__ = ("fragments", )

    def __init__(self):
        self.fragments = []

    def add(self, fragment):
        self.fragments.append(fragment)

    def get_string(self):
        result = "OPTIONAL { "
        for fragment in self.fragments:
            result = result + fragment.get_string() + " . "
        result += "}"
        return result


class Order(Node):
    __slots__ = ("by", "descending")

    def __init__(self, by, descending=False):
        self.by = by
        self.descending = descending

    def get_string(self):
        if self.descending:
            return "DESC(" + self.by.get_string() + ")"
        return self.by.get_string()


class PrefixedName(Symbol):
    __slots__ = ("name", )

    def __init__(self, name):
        self.name = name

    def get_string(self):
        return self.name


class SubQuery(object):
    __slots__ = ("at", "where", "query")

    def __init__(self, at, where, query):
        self.at = at
        self.where = where
        self.query = query


class Query(object):

    def __init__(self):
        self.selects = []
        self.where_fragments = []
        self.variables = {}

        self.subqueries = []
        self.order = []

        self.offset = 0
        self.limit = 0

    def add_order(self, by, descending):
        self.order.append(Order(by, descending))

    def add_variable(self, var):
        self.variables[var.name] = var

    def add(self, thing):
        if isinstance(thing, SelectFragment):
            self.selects.append(thing)
        elif isinstance(thing, WhereFragment):
            self.where_fragments.append(thing)
        else:
            raise ValueError("Can only add WhereFragment or SelectFragment to a Query")

    def add_subquery(self, at, where, query):
        self.subqueries.append(SubQuery(at, where, query));

    def has_variable(self, name):
        return name in self.variables

    def get_variable(self, name):
        return self.variables[name]

    def get_keys(self):
        for key in self.selects:
            yield key.get_name()

    def get_sparql(self):
        sparql = "SELECT "
        sparql += " ".join(s.get_string() for s in self.selects)
        sparql += " WHERE { "

        for f in self.where_fragments:
            if not isinstance(f, Optional):
                sparql += f.get_string() + " . "

        for f in self.where_fragments:
            if isinstance(f, Optional):
                sparql += f.get_string() + " . "

        sparql += " }"

        if len(self.order) > 0:
            sparql += " ORDER BY %s" % " ".join(o.get_string() for o in self.order)

        if self.offset > 0:
            sparql += " OFFSET " + unicode(self.offset)

        if self.limit > 0:
            sparql += " LIMIT " + unicode(self.limit)

        return sparql

    def get_model(self, async=True):
        return endpoint.get_results(self, async=async)


class Update(object):
    def get_sparql(self):
        raise NotImplementedError
    def execute(self):
        endpoint.update(self.get_sparql())


class Transaction(Update):
    def __init__(self, *args):
        self.updates = args
    def get_sparql(self):
        return " ".join(u.get_sparql() for u in self.updates)


class SparqlUpdate(Update):
    def __init__(self, sparql):
        self.sparql = sparql
    def get_sparql(self):
        return self.sparql


class Insert(Update):

    def __init__(self):
        self.relations = []

    def add(self, relationset):
        self.relations.append(relationset)

    def get_sparql(self):
        sparql = "INSERT { %s }" % " ".join(r.get_string() for r in self.relations)
        return sparql


class Delete(Update):

    def __init__(self):
        self.delete_triples = []
        self.where_triples = []

    def add_delete(self, triples):
        self.delete_triples.append(triples)

    def add_where(self, triples):
        self.where_triples.append(triples)

    def get_sparql(self):
        sparql = "DELETE { %s }" % " ".join(r.get_string() for r in self.delete_triples)
        if len(self.where_triples) > 0:
            sparql += " WHERE { %s }" % " ".join(r.get_string() for r in self.where_triples)
        return sparql

