# factory.py
#
# Copyright (C) 2009, Codethink Ltd.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#
# Author:
#       John Carr <john.carr@unrouted.co.uk>

import dbus

import dateutil.parser

from .resource import Resource
from .property import Property
from .utils import get_classname
from . import endpoint

# Map tracker prefixes to ontology namespaces and back
namespaces = {}
for prefix, namespace in endpoint.query("SELECT ?prefix ?ns WHERE { ?ns tracker:prefix ?prefix }"):
    namespaces[namespace] = prefix
del prefix, namespace


class Notifications(object):
    """ Class to allow users to attach signals to a class of object """
    def __init__(self, uri, signal):
        self.uri = uri.replace(":", "/")
        self.signal = signal

    def subscribe(self, callback):
        def _(subjects):
            subjects = [unicode(subject) for subject in subjects]
            return callback(subjects)
        bus.add_signal_receiver (_, signal_name=self.signal,
                                 dbus_interface="org.freedesktop.Tracker1.Resources.Class",
                                 path="/org/freedesktop/Tracker1/Resources/Classes/%s" % self.uri)


class WrapperFactory(object):

    def __init__(self):
        # Cache previous generated objects
        self.wrapped = {}

        # Connection to tracker
        self.bus = dbus.SessionBus()
        tracker_object = self.bus.get_object("org.freedesktop.Tracker1", "/org/freedesktop/Tracker1/Resources")
        self.tracker = dbus.Interface(tracker_object, "org.freedesktop.Tracker1.Resources")

        self.wrapped["xsd:boolean"] = lambda x: x=='true'
        self.wrapped["xsd:integer"] = int
        self.wrapped["xsd:double"] = float
        self.wrapped["rdfs:Literal"] = str
        self.wrapped["xsd:string"] = str
        self.wrapped['xsd:date'] = lambda x: dateutil.parser.parser().parse(x)
        self.wrapped['xsd:dateTime'] = lambda x: dateutil.parser.parser().parse(x)

    def get_class(self, classname):
        classname = get_classname(classname)

        if classname in self.wrapped:
            return self.wrapped[classname]

        ns, name = classname.split(":")

        sparql = "SELECT ?comment ?notify WHERE {" \
                     "%s a rdfs:Class . " \
                      "OPTIONAL { %s rdfs:comment ?comment } . " \
                      "OPTIONAL { %s tracker:notify ?notify } . " \
                 "}" % (classname, classname, classname)

        comment = None
        notify = None
        for comment, notify in endpoint.query(sparql):
            break

        attrs = {
            "_factory_": self,
            "_type_": classname,
            "__doc__": comment or "",
            "__module__": "tralchemy.%s" % ns,
        }

        baseclass = []
        sparql = "SELECT ?subclass WHERE { %s rdfs:subClassOf ?subclass }" % classname
        for value in endpoint.query(sparql):
            baseclass.append(self.get_class(unicode(value[0])))

        if not baseclass:
            baseclass.append(Resource)

        # An ontology might well have a pointless inheritenance chain
        # For example nmo:Message(TextDocument, InformationElement) - TextDocument is already an InfoEle!
        # Lets ignore the pointless inheritance and reduce chances of MRO related fail
        filtered = []
        for j in baseclass:
            for k in baseclass:
                if j != k and issubclass(k, j):
                    break
            else:
                filtered.append(j)

        # Does this class have notifications?
        if notify:
            attrs['added'] = Notifications(classname, 'SubjectsAdded')
            attrs['changed'] = Notifications(classname, 'SubjectsChanged')
            attrs['removed'] = Notifications(classname, 'SubjectsRemoved')

        sparql = "SELECT ?prop WHERE {" \
                     "?prop a rdf:Property ; " \
                          "rdfs:domain %s" \
                 "}" % classname

        for prop_uri in endpoint.query(sparql):
            prop_uri = get_classname(prop_uri[0])
            prop = Property(prop_uri, self)

            if len(prop_uri.split(":")) == 2:
                attrs[prop_uri.split(":")[1]] = prop
            elif len(prop_uri.split("#")) == 2:
                attrs[prop_uri.split("#")[1]] = prop

        # Make a new class
        klass = type(str(name), tuple(filtered), attrs)

        # Cache it for later
        self.wrapped[klass._type_] = klass

        return klass


types = WrapperFactory()

