# resource.py
#
# Copyright (C) 2009, Codethink Ltd.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#
# Author:
#       John Carr <john.carr@unrouted.co.uk>

import uuid

from . import endpoint
from .utils import get_classname
from .json import resource, query, insert, delete, transaction
from .query import Query


class Resource(object):
    """ Everything is a resource """

    _type_ = "rdfs:Resource"
    __module__ = "tralchemy.rdfs"

    def __init__(self, uri):
        self.raw_uri = uri
        self.uri = get_classname(uri)
        self.triples = {}
        self.cache = {}

    def properties(self):
        q = query({"resource": self.raw_uri, "?key": "?value"})
        for row in q.get_model(async=False):
            yield get_classname(unicode(row["key"])), unicode(row["value"])

    def iteritems(self):
        q = query({
            "name": "key",
            "type": "rdf:Property",
            "nrl:maxCardinality": {
                "name": "cardinality",
                "optional": True
                },
            "rdfs:range": "?range",
            "rdfs:domain": "?class",
            "badgers": [{
                "name": resource(self.raw_uri),
                "rdf:type": "?class",
                "?key": "?value",
                }],
            "order": ["key"],
            })

        def _(range, value):
            return self._factory_.get_class(range)(value)

        i = iter(q.get_model(async=False))
        row = i.next()
        while row:
            if row["cardinality"] == "1":
                yield get_classname(row["key"]), _(row["range"], row["value"])
                row = i.next()
            else:
                temp = [_(row["range"], row["value"])]
                key = row["key"]
                try:
                    row = i.next()
                    while row and row["key"] == key:
                        temp.append(_(row["range"], row["value"]))
                        row = i.next()
                    yield get_classname(key), temp
                except StopIteration as e:
                    yield get_classname(key), temp
                    raise e

    @classmethod
    def search(cls, search, **kwargs):
        q = {"name": "o",
             "type": cls._type_,
             "fts:match": search,
            }

        for key, value in kwargs.iteritems():
            if hasattr(value, "raw_uri"):
                value = resource(value.raw_uri)
            q[getattr(cls, key).uri] = value

        #FIXME pull this out of query
        rank = 0

        for result in query(q).get_model(async=False):
            yield cls(get_classname(result)), rank

    @classmethod
    def get(cls, **kwargs):
        q = {"type": cls._type_}
        for key, value in kwargs.iteritems():
            if isinstance(value, Query):
                value = value.json
            elif hasattr(value, "raw_uri"):
                value = resource(value.raw_uri)
            q[getattr(cls, key).uri] = value

        return Query(q, cls)

    @classmethod
    def create(cls, **kwargs):
        o = cls(kwargs.get('uid', 'urn:uuid:%s' % unicode(uuid.uuid4())))
        for k, v in kwargs.iteritems():
            if k == "uid":
                continue
            setattr(o, k, v)
        return o

    def delete(self):
        endpoint.update("DELETE { <%s> a rdfs:Resource }" % self.uri)

    def commit(self):
        self.get_transaction().execute()

    def get_transaction(self):
        steps = []
        for k, v in self.triples.iteritems():
            q = {
                "resource": self.raw_uri,
                "delete": True,
                k: "?v",
            }
            steps.append(delete(q))
        triples = {}
        triples["resource"] = self.raw_uri
        triples["type"] = self._type_

        for k, v in self.triples.iteritems():
            if hasattr(v, "uri"):
                 if v.dirty():
                      v.commit()
                 triples[k] = resource(v.raw_uri)
            elif isinstance(v, list):
                r = []
                for eep in v:
                    if hasattr(eep, "uri"):
                        if eep.dirty():
                            steps.append(eep.get_transaction())
                        r.append(resource(eep.raw_uri))
                    else:
                        r.append(eep)
                triples[k] = r
            else:
                triples[k] = v

        steps.append(insert(triples))
        self.triples = {}
        return transaction(*steps)

    def dirty(self):
        return len(self.triples) > 0
